# *** import

# ** core
from typing import Dict, Any

# ** infra
from tiferet.data import(
    DataObject,
    StringType,
    DictType,
    ModelType
)

# ** app
from ..models import (
    FlaskRoute, 
    FlaskBlueprint
)
from ..contracts import (
    FlaskRouteContract,
    FlaskBlueprintContract
)

# *** data

# ** data: flask_route_yaml_data
class FlaskRouteYamlData(DataObject, FlaskRoute):
    '''
    A data object for Flask route model from YAML.
    '''

    class Options():
        serialize_when_none = False
        roles = {
            'to_model': DataObject.allow(),
            'to_data': DataObject.deny('id')
        }

    # * attribute: id
    id = StringType(
        metadata=dict(
            description='The unique identifier of the route endpoint.'
        )
    )

    # * method: map
    def map(self, id: str) -> FlaskRouteContract:
        '''
        Map the data object to a FlaskRouteContract instance.

        :param id: The unique identifier of the route endpoint.
        :type id: str
        :return: A FlaskRouteContract instance.
        :rtype: FlaskRouteContract
        '''

        # Map the data object to a model instance.
        return super().map(
            FlaskRoute,
            id=id
        )

# ** data: flask_blueprint_yaml_data
class FlaskBlueprintYamlData(DataObject, FlaskBlueprint):
    '''
    A data object for Flask blueprint model from YAML.
    '''

    class Options():
        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('routes'),
            'to_data': DataObject.deny('name')
        }

    # * attribute: name
    name = StringType(
        metadata=dict(
            description='The name of the blueprint.'
        )
    )

    # * attribute: routes
    routes = DictType(
        ModelType(FlaskRouteYamlData),
        default={},
        metadata=dict(
            description='A dictionary of route ID to FlaskRouteYamlData instances.'
        )
    )

    # * method: from_data
    @staticmethod
    def from_data(routes: Dict[str, Any] = {}, **kwargs) -> 'FlaskBlueprintYamlData':
        '''
        Create a FlaskBlueprintYamlData instance from raw data.

        :param routes: A dictionary of route ID to route data.
        :type routes: Dict[str, Any]
        :return: A FlaskBlueprintYamlData instance.
        :rtype: FlaskBlueprintYamlData
        '''

        # Convert each route in the routes dictionary to a FlaskRouteYamlData instance.
        route_objs = {id: DataObject.from_data(
            FlaskRouteYamlData, 
            id=id, **data
        ) for id, data in routes.items()}

        # 
        return DataObject.from_data(
            FlaskBlueprintYamlData,
            routes=route_objs,
            **kwargs
        )

    # * method: map
    def map(self) -> FlaskBlueprintContract:
        '''
        Map the data object to a FlaskBlueprintContract instance.

        :return: A FlaskBlueprintContract instance.
        :rtype: FlaskBlueprintContract
        '''

        # Map each route in the routes dictionary.
        return super().map(
            FlaskBlueprint,
            routes=[route.map(id=id) for id, route in self.routes.items()]
        )


# *** imports

# ** core
from typing import List

# ** infra
from tiferet import (
    ModelContract,
    Repository,
    abstractmethod
)

# *** contracts

# ** contract: flask_route_contract
class FlaskRouteContract(ModelContract):
    '''
    A contract for Flask route models.
    '''

    # * attribute: id
    id: str

    # * attribute: status_code
    status_code: int

# ** contract: flask_blueprint_contract
class FlaskBlueprintContract(ModelContract):
    '''
    A contract for Flask blueprint models.
    '''

    # * attribute: name
    name: str

    # * attribute: routes
    routes: List[FlaskRouteContract]

# ** contract: flask_api_repository
class FlaskApiRepository(Repository):
    '''
    A repository contract for managing Flask API entities.
    '''

    # * method: get_blueprints
    @abstractmethod
    def get_blueprints(self) -> List[FlaskBlueprintContract]:
        '''
        Retrieve all Flask blueprints.

        :return: A list of FlaskBlueprintContract instances.
        :rtype: List[FlaskBlueprintContract]
        '''
        raise NotImplementedError('get_blueprints method not implemented.')

    # * method: get_route
    @abstractmethod
    def get_route(self, route_id: str, blueprint_name: str = None) -> FlaskRouteContract:
        '''
        Retrieve a specific Flask route by its blueprint and route IDs.

        :param route_id: The ID of the route within the blueprint.
        :type route_id: str
        :param blueprint_name: The name of the blueprint (optional).
        :type blueprint_name: str
        :return: The corresponding FlaskRouteContract instance.
        :rtype: FlaskRouteContract
        '''
        raise NotImplementedError('get_route method not implemented.')

    # * method: get_status_code
    @abstractmethod
    def get_status_code(self, error_code: str) -> int:
        '''
        Retrieve the HTTP status code for a given error code.

        :param error_code: The error code.
        :type error_code: str
        :return: The corresponding HTTP status code.
        :rtype: int
        '''
        raise NotImplementedError('get_status_code method not implemented.')


# *** imports

# ** core
from typing import List, Any

# ** infra
from tiferet import (
    TiferetError, 
    raise_error
)
from tiferet.proxies.yaml import (
    YamlConfigurationProxy
)

# ** app
from ...contracts import (
    FlaskApiRepository,
    FlaskBlueprintContract,
    FlaskRouteContract
)
from ...data import (
    FlaskBlueprintYamlData,
)

# *** proxies

# ** proxy: flask_yaml_proxy
class FlaskYamlProxy(FlaskApiRepository, YamlConfigurationProxy):
    '''
    A YAML configuration proxy for Flask settings.
    '''

    # * init
    def __init__(self, flask_config_file: str):
        '''
        Initialize the FlaskYamlProxy with the given YAML file path.

        :param flask_config_file: The path to the Flask configuration YAML file.
        :type flask_config_file: str
        '''

        # Set the configuration file within the base class.
        super().__init__(flask_config_file)

     # * method: load_yaml
    def load_yaml(self, start_node: callable = lambda data: data, create_data: callable = lambda data: data) -> Any:
        '''
        Load data from the YAML configuration file.
        :param start_node: The starting node in the YAML file.
        :type start_node: str
        :param create_data: A callable to create data objects from the loaded data.
        :type create_data: callable
        :return: The loaded data.
        :rtype: Any
        '''

        # Load the YAML file contents using the yaml config proxy.
        try:
            return super().load_yaml(
                start_node=start_node,
                create_data=create_data
            )

        # Raise an error if the loading fails.
        except (Exception, TiferetError) as e:
            raise_error.execute(
                'FLASK_CONFIG_LOADING_FAILED',
                f'Unable to load flask configuration file {self.config_file}: {e}.',
                self.config_file,
                str(e)
            )

    # * method: get_blueprints
    def get_blueprints(self) -> List[FlaskBlueprintContract]:
        '''
        Retrieve all Flask blueprints from the YAML configuration.

        :return: A list of FlaskBlueprintContract instances.
        :rtype: List[FlaskBlueprintContract]
        '''

        # Load the blueprints section from the YAML file.
        data = self.load_yaml(
            create_data=lambda data: [FlaskBlueprintYamlData.from_data(
                name=name,
                **blueprint
            ) for name, blueprint in data.items()],
            start_node=lambda d: d.get('flask', {}).get('blueprints', {})
        )

        # Map the loaded data to FlaskBlueprintContract instances.
        return [blueprint.map() for blueprint in data]

    # * method: get_route
    def get_route(self, route_id: str, blueprint_name: str = None) -> FlaskRouteContract:
        '''
        Retrieve a specific Flask route by its blueprint and route IDs from the YAML configuration.

        :param route_id: The route identifier.
        :type route_id: str
        :param blueprint_name: The name of the blueprint (optional).
        :type blueprint_name: str
        :return: The corresponding FlaskRouteContract instance.
        :rtype: FlaskRouteContract
        '''

        # Load the blueprints section from the YAML file.
        blueprints = self.get_blueprints()

        # Search for the specified blueprint.
        for blueprint in blueprints:
            if blueprint_name and blueprint.name != blueprint_name:
                continue

            # Search for the route within the blueprint.
            for route in blueprint.routes:
                if route.id == route_id:
                    return route

        # If not found, return None.
        return None

    # * method: get_status_code
    def get_status_code(self, error_code: str) -> int:
        '''
        Retrieve the HTTP status code for a given error code from the YAML configuration.

        :param error_code: The error code identifier.
        :type error_code: str
        :return: The corresponding HTTP status code.
        :rtype: int
        '''

        # Load the error code from the errors section of the YAML file.
        data = self.load_yaml(
            start_node=lambda d: d.get('flask').get('errors', {}),
        )

        # Return the status code if found, otherwise default to 500.
        return data.get(error_code, 500)


# *** imports

# ** core
from typing import List

# ** infra
from tiferet.commands import raise_error

# ** app
from ..contracts import (
    FlaskBlueprintContract,
    FlaskRouteContract,
    FlaskApiRepository
)

# *** handlers

# ** handler: flask_api_handler
class FlaskApiHandler(object):
    '''
    A handler for managing Flask API entities.
    '''

    # * init
    def __init__(self, flask_repo: FlaskApiRepository):
        '''
        Initialize the FlaskHandler with the given repository.

        :param flask_repo: An instance of FlaskApiRepository.
        :type flask_repo: FlaskApiRepository
        '''

        # Store the repository instance.
        self.flask_repo = flask_repo

    # * method: get_blueprints
    def get_blueprints(self) -> List[FlaskBlueprintContract]:
        '''
        Retrieve all Flask blueprints using the repository.

        :return: A list of FlaskBlueprintContract instances.
        :rtype: List[FlaskBlueprintContract]
        '''

        # Delegate the call to the repository.
        return self.flask_repo.get_blueprints()

    # * method: get_route
    def get_route(self, endpoint: str) -> FlaskRouteContract:
        '''
        Retrieve a specific Flask route by its blueprint and route IDs.

        :param endpoint: The endpoint in the format 'blueprint_name.route_id'.
        :type endpoint: str
        :return: The corresponding FlaskRouteContract instance.
        :rtype: FlaskRouteContract
        '''

        # Split the endpoint into blueprint and route IDs.
        # If no blueprint is specified, assume the route is at the root level.
        blueprint_name = None
        try:
            blueprint_name, route_id = endpoint.split('.')
        except ValueError:
            route_id = endpoint

        # Delegate the call to the repository.
        route = self.flask_repo.get_route(
            route_id=route_id,
            blueprint_name=blueprint_name,
        )

        # Raise an error if the route is not found.
        if route is None:
            raise_error.execute(
                'FLASK_ROUTE_NOT_FOUND',
                f'Flask route not found for endpoint: {endpoint}',
                endpoint
            )

        # Return the found route.
        return route

    # * method: get_status_code
    def get_status_code(self, error_code: str) -> int:
        '''
        Retrieve the HTTP status code for a given error code using the repository.

        :param error_code: The error code identifier.
        :type error_code: str
        :return: The corresponding HTTP status code.
        :rtype: int
        '''

        # Delegate the call to the repository.
        return self.flask_repo.get_status_code(
            error_code=error_code
        )


# *** imports

# ** core
from typing import Any

# ** infra
from tiferet.contexts.request import RequestContext
from tiferet.models import ModelObject

# *** contexts

# ** context: flask_request_context
class FlaskRequestContext(RequestContext):

    # * method: handle_response
    def handle_response(self) -> Any:
        '''
        Handle the response for the Flask request context.

        :return: The response.
        :rtype: Any
        '''

        # Set the result using the set_result method to ensure proper formatting.
        self.set_result(self.result)

        # Handle the response using the parent method.
        return super().handle_response()

    # * method set_result
    def set_result(self, result: Any):
        '''
        Set the result of the request context.

        :param result: The result to set.
        :type result: Any
        '''

        # If the response is None, return an empty response.
        if result is None:
            self.result = ''

        # Convert the response to a dictionary if it's a ModelObject.
        elif isinstance(result, ModelObject):
            self.result = result.to_primitive()

        # If the response is a list containing model objects, convert each to a dictionary.
        elif isinstance(result, list) and all(isinstance(item, ModelObject) for item in result):
            self.result = [item.to_primitive() for item in result]

        # If the rsponse is a dict containing model objects, convert each to a dictionary.
        elif isinstance(result, dict) and all(isinstance(value, ModelObject) for value in result.values()):
            self.result = {key: value.to_primitive() for key, value in result.items()}

        # Otherwise, set the result directly.
        else:
            self.result = result




# *** imports

# ** core
from typing import Any, Callable

# ** infra
from flask import Flask, Blueprint
from tiferet.contexts.app import (
    AppInterfaceContext, 
    RequestContext,
    TiferetError
)
from tiferet.contexts.request import RequestContext
from tiferet.contexts.error import ErrorContext
from tiferet.contexts.feature import FeatureContext
from tiferet.contexts.logging import LoggingContext

# ** app
from .request import FlaskRequestContext
from ..handlers.flask import FlaskApiHandler
from ..models import FlaskBlueprint

# *** contexts

# ** context: flask_api_context
class FlaskApiContext(AppInterfaceContext):
    '''
    A context for managing Flask API interactions within the Tiferet framework.
    '''

    # * attribute: flask_app
    flask_app: Flask

    # * attribute: flask_api_handler
    flask_api_handler: FlaskApiHandler

    # * init
    def __init__(self, 
        interface_id: str,
        features: FeatureContext,
        errors: ErrorContext,
        logging: LoggingContext,
        flask_api_handler: FlaskApiHandler
    ):
        '''
        Initialize the application interface context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param features: The feature context.
        :type features: FeatureContext
        :param errors: The error context.
        :type errors: ErrorContext
        :param logging: The logging context.
        :type logging: LoggingContext
        :param flask_api_handler: The Flask API handler.
        :type flask_api_handler: FlaskApiHandler
        '''

        # Call the parent constructor.
        super().__init__(interface_id, features, errors, logging)

        # Set the attributes.
        self.flask_api_handler = flask_api_handler

    # * method: parse_request
    def parse_request(self, headers: dict = {}, data: dict = {}, feature_id: str = None, **kwargs) -> FlaskRequestContext:
        '''
        Parse the incoming request and return a FlaskRequestContext instance.

        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :param feature_id: The feature ID.
        :type feature_id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A FlaskRequestContext instance.
        :rtype: FlaskRequestContext
        '''

        # Return a FlaskRequestContext instance.
        return FlaskRequestContext(
            headers=headers,
            data=data,
            feature_id=feature_id,
        )

    # * method: handle_error
    def handle_error(self, error: Exception) -> Any:
        '''
        Handle the error and return the response.

        :param error: The error to handle.
        :type error: Exception
        :return: The error response.
        :rtype: Any
        '''

        # Handle the error and get the response from the parent context.
        if not isinstance(error, TiferetError):
            return super().handle_error(error), 500

        # Get the status code by the error code on the exception.
        status_code = self.flask_api_handler.get_status_code(error.error_code)
        return super().handle_error(error), status_code

    # * method: handle_response
    def handle_response(self, request: RequestContext) -> Any:
        '''
        Handle the response from the request context.

        :param request: The request context.
        :type request: RequestContext
        :return: The response.
        :rtype: Any
        '''

        # Handle the response from the request context.
        response = super().handle_response(request)

        # Retrieve the route by the request feature id.
        route = self.flask_api_handler.get_route(request.feature_id)

        # Return the result as JSON with the specified status code.
        return response, route.status_code

    # * method: build_blueprint
    def build_blueprint(self, flask_blueprint: FlaskBlueprint, view_func: Callable, **kwargs) -> Blueprint:
        '''
        Assembles a Flask blueprint from the given FlaskBlueprint model.

        :param flask_blueprint: The FlaskBlueprint model.
        :type flask_blueprint: FlaskBlueprint
        :param view_func: The view function to handle requests.
        :type view_func: Callable
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The created Flask blueprint.
        :rtype: Blueprint
        '''

        # Create the blueprint.
        blueprint = Blueprint(
            flask_blueprint.name, 
            __name__, 
            url_prefix=flask_blueprint.url_prefix
        )

        # Add the url rules.
        for route in flask_blueprint.routes:
            blueprint.add_url_rule(
                route.rule, 
                route.id, 
                methods=route.methods, 
                view_func=lambda: view_func(self, **kwargs),
            )

        # Return the created blueprint.
        return blueprint

    # * method: build_flask_app
    def build_flask_app(self, view_func: Callable, **kwargs) -> Flask:
        '''
        Build and return a Flask application instance.

        :param view_func: The view function to handle requests.
        :type view_func: Callable
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A Flask application instance.
        :rtype: Flask
        '''

        # Import CORS here to avoid circular import issues.
        from flask_cors import CORS

        # Create the Flask application.
        # Enable CORS for the Flask application.
        flask_app = Flask(__name__)
        CORS(flask_app)

        # Load the Flask blueprints.
        blueprints = self.flask_api_handler.get_blueprints()

        # Create and register the blueprints.
        for bp in blueprints:
            blueprint = self.build_blueprint(bp, view_func=view_func, **kwargs)
            flask_app.register_blueprint(blueprint)

        # Set the flask_app attribute.
        self.flask_app = flask_app


# *** imports

# ** infra
from tiferet.models import (
    ModelObject,
    StringType,
    IntegerType,
    ListType,
    ModelType
)

# *** models

# ** model: flask_route
class FlaskRoute(ModelObject):
    '''
    A Flask route model.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier of the route endpoint.'
        )
    )

    # * attribute: rule
    rule = StringType(
        required=True,
        metadata=dict(
            description='The URL rule as string.'
        )
    )

    # * attribute: methods
    methods = ListType(
        StringType,
        required=True,
        metadata=dict(
            description='A list of HTTP methods this rule should be limited to.'
        )
    )

    # * attribute: status_code
    status_code = IntegerType(
        default=200,
        metadata=dict(
            description='The default HTTP status code for the route response.'
        )
    )

# ** model: flask_blueprint
class FlaskBlueprint(ModelObject):
    '''
    A Flask blueprint model.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the blueprint.'
        )
    )

    # * attribute: url_prefix
    url_prefix = StringType(
        metadata=dict(
            description='The URL prefix for all routes in this blueprint.'
        )
    )

    # * attribute: routes
    routes = ListType(
        ModelType(FlaskRoute),
        default=[],
        metadata=dict(
            description='A list of routes associated with this blueprint.'
        )
    )